try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

from scipy import signal

__all__ = ['OMFITtimcon']


class OMFITtimcon(SortedDict, OMFITascii):
    """
    OMFIT class for parsing/writing the NBI time traces of the DIII-D PCS
    """

    def __init__(self, filename):
        OMFITascii.__init__(self, filename)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as _f:
            lines = _f.read().split('\n')
        nummod = int(lines[0].split('=')[1])
        numphases = int(lines[1].split('=')[1])
        k = 2
        mode_lookup = {0: 'NoMod', 1: 'PCS', 2: 'TIMCON', 3: 'CALORIMETER', 4: r'PCS/TIMCON'}
        states_lookup = {
            0: 'PRI:AnyGen',
            1: 'PRI:LikeSub',
            2: 'PRI:NoSub',
            3: 'PRI:SpecialA',
            4: 'PRI:SpecialB',
            5: 'PRI:SpecialC',
            6: 'SEC:Gen/Like',
            7: 'SEC:SpecialA',
            8: 'SEC:SpecialB',
            9: 'SEC:SpecialC',
        }
        for mod in range(nummod):
            tmp = lines[k].split('-')
            modname = tmp[0].strip().split()[1]
            self[modname] = SortedDict()
            tmp = tmp[1].split()
            self[modname]['enabled'] = bool(int(tmp[0]))
            self[modname]['total_on_time'] = float(tmp[1])
            self[modname]['mode'] = int(tmp[2])
            self[modname]['mode_txt'] = mode_lookup[self[modname]['mode']]
            self[modname]['beam_status'] = int(tmp[3])
            self[modname]['beam_status_txt'] = states_lookup[self[modname]['beam_status']]
            self[modname]['sub_beam'] = self[modname]['mode_txt'] == 'PCS' and self[modname]['beam_status_txt'].find('SEC') == 0
            self[modname]['sub_priority'] = int(tmp[4])
            self[modname]['__unknown__'] = int(tmp[5])
            k += 1
            for phase in range(1, numphases + 1):
                tmp = list(map(eval, lines[k].split(':')[1].split()))
                self[modname]['PHASE%d' % phase] = SortedDict()
                for k1, item in enumerate(['__enabled__', 'start', 'on', 'off', 'duration', 'pulses']):
                    self[modname]['PHASE%d' % phase][item] = tmp[k1]
                # note the timcon files generated by timcon always have the enabled phase switched turned on (that's why we hide it here)
                self[modname]['PHASE%d' % phase]['__enabled__'] = bool(self[modname]['PHASE%d' % phase]['__enabled__'])
                k += 1

    @dynaSave
    def save(self):
        lines = []
        lines.append('nummod = %d' % len(self))
        numphases = 100
        for modname in self:
            numphases = min([numphases, len([phase for phase in self[modname] if phase.startswith('PHASE')])])
        lines.append('numphases = %d' % numphases)
        for kmod, modname in enumerate(self):
            lines.append(
                (
                    'SRC%d %s - {enabled:d} {total_on_time:0.6f} {mode:d} {beam_status:d} {sub_priority:d} {__unknown__:d}'
                    % (kmod + 1, modname)
                ).format(**self[modname])
            )
            for phase in self[modname]:
                if not phase.startswith('PHASE'):
                    continue
                lines.append(
                    (
                        '%s: %d %d %d %d %d %0.6f'
                        % (
                            phase,
                            self[modname][phase]['__enabled__'],
                            self[modname][phase]['start'],
                            self[modname][phase]['on'],
                            self[modname][phase]['off'],
                            self[modname][phase]['duration'],
                            self[modname][phase]['pulses'],
                        )
                    )
                )
        with open(self.filename, 'w') as f:
            f.write('\n'.join(lines) + '\n')

    @dynaLoad
    def waveform(self, beam, x, y=None, phases=None, duty_cycle_1=False):
        if phases is None:
            phases = [phase for phase in self[beam] if phase.startswith('PHASE')]
        else:
            phases = tolist(phases)
        if y is None:
            y = x * 0

        starts = np.array([self[beam][phase]['start'] for phase in phases])
        durations = np.array([self[beam][phase]['duration'] for phase in phases])
        ons = np.array([self[beam][phase]['on'] for phase in phases])
        ons *= np.array([self[beam]['enabled'] * self[beam][phase]['__enabled__'] for phase in phases])
        offs = np.array([self[beam][phase]['off'] for phase in phases])

        for k in range(len(phases)):
            x0 = starts[k]
            dur = durations[k]
            on1 = ons[k]
            off1 = offs[k]

            if on1 == 0:
                # abort if no on time
                continue

            # calculations
            x1 = x0 + dur
            if duty_cycle_1:
                period = 100
                duty = 1.0
            else:
                period = on1 + off1
                duty = on1 / float(period)

            # form the output
            w = np.where((x > x0) * (x < x1))
            tt = (x[w] - x0) * 2 * np.pi / period
            y[w] = (signal.square(tt, duty=duty) + 1.0) / 2.0

        return x, y

    @dynaLoad
    def plot_waveform(self, beam, ax=None, show_legend=True):
        x = np.linspace(0, 10000, 10000)
        if ax is None:
            ax = pyplot.gca()
        x, y = self.waveform(beam, x)
        color_blind_color_cycle = [
            '#1f77b4',
            '#ff7f0e',
            '#2ca02c',
            '#d62728',
            '#9467bd',
            '#8c564b',
            '#e377c2',
            '#7f7f7f',
            '#bcbd22',
            '#17becf',
        ]
        label = f"{self[beam]['mode_txt']} {self[beam]['beam_status_txt']}"
        ax.plot(x, y, color=color_blind_color_cycle[self[beam]['mode']], label=label)
        ax.set_ylim([0, 2])
        ax.set_ylabel(beam)
        for kphase, phase in enumerate(self[beam]):
            if not phase.startswith('PHASE'):
                continue
            ax.axvspan(
                self[beam][phase]['start'],
                self[beam][phase]['start'] + self[beam][phase]['duration'],
                color=['lightcoral', 'turquoise'][np.mod(kphase, 2)],
                alpha=0.15,
                linewidth=0,
            )
        ax.set_yticks([0, 1])
        ax.set_xlabel('Time (ms)')
        if show_legend:
            ax.legend(loc=0, frameon=False, text_same_color=True, hide_markers=True, fontsize='small')
        return x, y

    @dynaLoad
    def plot(self):
        x = np.linspace(0, 10000, 10000)
        y = x * 0
        nrow = len(self) + 1
        # what magic is making this method create a new figure?
        # it is not transparent how I a) pass an existing figure for over-plotting, b) specify the size :(
        fig = pyplot.gcf()
        fig.set_size_inches(8, 9)
        ax = pyplot.subplot(nrow, 1, 1)
        ax.set_xlim(x[0], x[-1])  # have to get the x10^4 label in now so the next axes hide it
        ax.set_ylabel('ALL')
        ax.set_ylim(0, len(self) + 2)
        for i, beam in enumerate(self):
            ai = pyplot.subplot(nrow, 1, i + 2, sharex=ax)
            x, y0 = self.plot_waveform(beam, ax=ai)
            y += y0
        ax.plot(x, y, 'k')
        autofmt_sharex(hspace=0.1)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import numpy as np
import scipy as scipy
from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITcoils', 'OMFITfocuscoils', 'OMFITGPECcoils']


class OMFITcoils(SortedDict, OMFITascii):
    """
    OMFITobject used to interface with coils ascii files used in FOCUS and STELLOPT codes.

    :param filename: filename passed to OMFITascii class

    All additional key word arguments passed to OMFITascii

    """

    def __init__(self, filename, **kwargs):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kwargs)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        """Load the file and parse it into a sorted dictionary"""
        with open(self.filename) as f:
            lines = f.readlines()

        # allow class to be initialized with a new (empty) file
        if len(lines) == 0:
            return

        # first three lines are global parameters
        # only the first line has meanings
        for i in range(1):
            k, v = lines[i].split()
            if v.isdigit():
                self[k] = int(v)
            else:
                self[k] = v

        self['num_coils'] = 0

        # find end of data
        nlines = len(lines)
        end_line = 3
        for i in range(-1, -nlines, -1):
            if lines[i].lstrip().lower().startswith('end'):
                end_line = nlines + i
                break

        # if no 'end' detected, use the last line
        if end_line == 3:
            end_line = nlines

        # find the start, stop of each coil
        start = 3
        check = 0
        num_coil = 0
        while start < end_line:
            # defend against runaway while loops
            check += 1
            if check > 1e6:
                raise RuntimeError("Found over 1e6 coils. That is too many!")

            # read each coil
            xyzc = []
            for i in range(start, end_line):
                xyzc.append(list(map(np.float, lines[i].split()[:4])))
                if len(lines[i].split()) != 4:  # detect new coil
                    group, name = lines[i].split()[4:6]
                    num_coil += 1
                    # assign labels for each coil, even they have the same names
                    if name not in self:
                        label = name  # default name
                    else:
                        for icoil in range(1, 10001):
                            label = '{:}_{:}'.format(name, icoil)
                            if label not in self:
                                break
                            if icoil == 10000:
                                raise OMFITexception('Only 10000 or fewer coils can have the same name.')

                    xyzc = np.array(xyzc)
                    self[label] = SortedDict()
                    self[label]['x'] = xyzc[:, 0]
                    self[label]['y'] = xyzc[:, 1]
                    self[label]['z'] = xyzc[:, 2]
                    self[label]['I'] = xyzc[0, 3]  # coil current is just a float value
                    self[label]['group'] = int(group)
                    self[label]['name'] = name
                    start = i + 1
                    break
        self['num_coils'] = num_coil  # total number of coils
        return

    @dynaSave
    def save(self):
        """Write the data in standard ascii format"""
        with open(self.filename, 'w') as f:
            # write the defining parameters
            periods = list(self.keys())[0]

            f.write('{:} {:}\n'.format(periods, self[periods]))  # the first line with periods
            f.write('{:} {:}\n'.format('begin', 'filament'))
            f.write('{:} {:}\n'.format('mirror', 'NIL'))

            # write each coil
            for key, coil in list(self.items()):
                if not isinstance(coil, dict):
                    continue
                for i in range(len(coil['x']) - 1):
                    f.write('{:23.15E}{:23.15E}{:23.15E}{:23.15E}\n'.format(coil['x'][i], coil['y'][i], coil['z'][i], coil['I']))
                i = len(coil['x']) - 1
                f.write(
                    '{:23.15E}{:23.15E}{:23.15E}{:23.15E} {:>11} {:>11}\n'.format(
                        coil['x'][i], coil['y'][i], coil['z'][i], 0.0, coil['group'], coil['name']
                    )
                )
            f.write('end\n')
        return

    @dynaLoad
    def plot(self, ax=None, nd=3, cmap='coolwarm_r', vmin=None, vmax=None, colorbar=True, **kwargs):
        r"""
        Plot the coils in 3D, colored by their current values.

        :param ax: Axes. Axis into which the plot will be made.

        :param cmap: string. A valid matplolib color map name.

        :param vmin: float. Minimum of the color scaling.

        :param vmax: float. Maximum of the color scaling.

        :param colorbar: bool. Draw a colorbar for the currents.

        :param \**kwargs: dict. All other key word arguments are passed to the mplot3d plot function.

        :return: list of Line3D objects. The line plotted for each coil.

        """
        if nd == 3:
            projection = '3d'
        else:
            projection = None
        if ax is None:
            # extra logic to handle OMFIT's auto-figure making when double-clicked in tree
            f = pyplot.gcf()
            if len(f.axes):
                # normal situation in which existing figures should be respected and left alone
                try:
                    f, ax = pyplot.subplots(subplot_kw={'aspect': 'equal', 'projection': projection})
                except NotImplementedError:  # python3 matplolib did not have this for some reason
                    pyplot.close(pyplot.gcf())  # will have opened a figure than failed to set the aspect
                    f, ax = pyplot.subplots(subplot_kw={'projection': projection})
            else:
                # OMFIT (or the gcf needed to check OMFIT) made a empty figure for us
                ax = f.add_subplot(111, projection=projection)  # OKadd
                try:
                    ax.set_aspect('equal')
                except NotImplementedError:
                    pass
            # aesthetics
            if nd == 3:
                ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
                ax.set_xlabel('x (m)')
                ax.set_ylabel('y (m)')
                ax.set_zlabel('z (m)')
                ax.set_axis_off()
        else:
            f = ax.get_figure()

        # defaults
        kwargs.setdefault('linewidth', 3)
        if 'lw' in kwargs:
            kwargs['linewidth'] = kwargs.pop('lw')

        keys = list(self.keys())
        curs, lines = [], []
        bound = 0
        for key, coil in list(self.items()):
            if not isinstance(coil, dict):
                continue
            curs.append(coil['I'])
            if nd == 3:
                lines.append(ax.plot(coil['x'], coil['y'], coil['z'], label=key, **kwargs)[0])
                bound = max(bound, np.max(np.abs(np.hstack((coil['x'], coil['y'], coil['z'])))))
            elif nd == 1:
                r = np.sqrt(coil['x'] ** 2 + coil['y'] ** 2)
                lines.append(ax.plot(r, coil['z'], label=key, **kwargs)[0])
                bound = max(bound, np.max(np.abs(np.hstack((r, coil['z'])))))
            else:
                raise OMFITexception("Only 3D and 1D plots are available")
        # color by current
        from omfit_classes.utils_plot import set_linearray

        sm = set_linearray(lines, values=curs, vmin=vmin, vmax=vmax, cmap=cmap)
        if colorbar:
            cb = f.colorbar(sm, ax=ax, shrink=0.8)
            cb.set_label('Current (A)')

        # force equal aspect
        if nd == 3:
            bound *= 0.6
            ax.set_xlim(-bound, bound)
            ax.set_ylim(-bound, bound)
            ax.set_zlim(-bound, bound)
            ax.auto_scale_xyz((-bound, bound), (-bound, bound), (-bound, bound))

        return lines

    @dynaLoad
    def to_OMFITfocuscoils(self, filename, nfcoil=8, target_length=0, coil_type=1, coil_symm=0, nseg=None, i_flag=1, l_flag=0):
        """
        Convert to the  OMFITfocuscoils ascii file format, which stores Fourier Harmonics
        of the coils. These files have additional settings, which can be set using key
        word arguments. Be sure to sync these with the focus input namelist!

        :param filename: str. Name of new file

        :param nfcoil: int. Number of harmonics in decomposition

        :param target_length: float. Target length of coils. Zero targets the initial length.

        :param coil_type: int.

        :param coil_symm: int. 0 for no symmetry. 1 for toroidal symmetry matching the boundary bnfp.

        :param nseg: int. Number of segments. Default (None) uses the number of segments in the original file

        :param i_flag: int.

        :param l_flag: int.

        :return: OMFITcoils

        """
        coils = SortedDict()
        ckeys = [k for k, v in self.items() if isinstance(v, dict)]
        for i, key in enumerate(ckeys):
            # decompose x,y,z
            A = []
            theta = np.linspace(0, 2 * np.pi, len(self[key]['x']))
            for t in theta:
                A.append(np.ravel([(np.cos(m * t), np.sin(m * t)) for m in range(nfcoil + 1)]))
            A = np.array(A)
            csx = scipy.linalg.lstsq(A, self[key]['x'], 1e-2)[0]
            csy = scipy.linalg.lstsq(A, self[key]['y'], 1e-2)[0]
            csz = scipy.linalg.lstsq(A, self[key]['z'], 1e-2)[0]
            length = np.sum(np.sqrt(np.diff(self[key]['x']) ** 2 + np.diff(self[key]['y']) ** 2 + np.diff(self[key]['z']) ** 2))
            # fill in info
            coils[key] = SortedDict()
            coils[key]['coil_type'] = coil_type  # todo: learn what this is
            coils[key]['coil_symm'] = coil_symm
            if nseg is None:
                coils[key]['nseg'] = len(self[key]['x']) - 1
            else:
                coils[key]['nseg'] = nseg
            coils[key]['current'] = self[key]['I']
            coils[key]['i_flag'] = i_flag
            coils[key]['length'] = length
            coils[key]['l_flag'] = l_flag
            if target_length < 0:
                coils[key]['target_length'] = length
            else:
                coils[key]['target_length'] = target_length
            coils[key]['nfcoil'] = nfcoil
            coils[key]['xc'] = csx[0::2]
            coils[key]['xs'] = csx[1::2]
            coils[key]['yc'] = csy[0::2]
            coils[key]['ys'] = csy[1::2]
            coils[key]['zc'] = csz[0::2]
            coils[key]['zs'] = csz[1::2]

        # Use the OMFIT class to create an empty file and then update it in memory
        result = OMFITfocuscoils(filename)
        result.update(coils)

        return result

    @dynaLoad
    def to_OMFITGPECcoils(self, filename):
        """
        Convert OMFITcoils to the OMFITGPECcoils ascii file format, which writes x,y,z points
        but no current information.

        :param filename: str. Name of new file

        :return: OMFITGPECcoils

        """
        # Use the OMFIT class to create an empty file and then update it in memory
        result = OMFITGPECcoils(filename)
        result.update(self)

        return result


class OMFITfocuscoils(SortedDict, OMFITascii):
    """
    OMFITobject used to interface with focus ascii files used to describe coils in native FOCUS
    decomposition.

    :param filename: filename passed to OMFITascii class

    All additional key word arguments passed to OMFITascii

    """

    def __init__(self, filename, **kwargs):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kwargs)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        """Load the file and parse it into a sorted dictionary"""
        with open(self.filename) as f:
            lines = f.readlines()

        # allow class to be initialized with a new (empty) file
        if len(lines) == 0:
            return

        ncoils = int(lines[1].strip())

        for i in range(ncoils):
            start = 2 + i * 14
            ctype, csym, cname = lines[start + 2].split()
            self[cname] = SortedDict()
            self[cname]['coil_type'] = int(ctype)
            self[cname]['coil_symm'] = int(csym)
            keys = lines[start + 3].split()[1:]
            vals = lines[start + 4].split()
            self[cname]['nseg'] = int(vals[0])
            self[cname]['current'] = float(vals[1])
            self[cname]['i_flag'] = int(vals[2])
            self[cname]['length'] = float(vals[3])
            self[cname]['l_flag'] = int(vals[4])
            self[cname]['target_length'] = float(vals[5])
            self[cname]['nfcoil'] = int(lines[start + 6].strip())
            for j, key in enumerate(['xc', 'xs', 'yc', 'ys', 'zc', 'zs']):
                self[cname][key] = np.array(lines[start + 8 + j].split(), dtype=float)
        return

    @dynaSave
    def save(self):
        """Write the data in standard ascii format"""
        with open(self.filename, 'w') as f:
            # write the total number
            f.write(' # ncoils\n')
            f.write('{:12}\n'.format(len(self)))
            for i, key in enumerate(self.keys()):
                f.write(' #--------------------------------------------\n')
                f.write(' #   coil_type    coil_symm    coil_name\n')
                f.write('  {:12} {:12} {:>12}\n'.format(self[key]['coil_type'], self[key]['coil_symm'], key))
                keys = ['nseg', 'current', 'i_flag', 'length', 'l_flag', 'target_length']
                vals = [self[key][k] for k in keys]
                f.write(' # {:>6} {:>18} {:>6} {:>18} {:>6} {:>18}\n'.format(*keys))
                f.write('   {:6} {:18.9E} {:6} {:18.9E} {:6} {:18.9E}\n'.format(*vals))
                f.write(' #      nfcoil\n')
                f.write('  {:12}\n'.format(self[key]['nfcoil']))
                f.write(' # Fourier harmonics for coils ( xc; xs; yc; ys; zc; zs)\n')
                for k in ['xc', 'xs', 'yc', 'ys', 'zc', 'zs']:
                    f.write((' {:23.15E}' * len(self[key][k]) + '\n').format(*self[key][k]))
        return

    @dynaLoad
    def plot(self, bnfp=1, ax=None, cmap='coolwarm_r', vmin=None, vmax=None, colorbar=True, **kwargs):
        r"""
        Plot the coils in 3D, colored by their current values.

        :param bnfp: int. Toroidal mode number of symmetry used if coil_symm is 1

        :param ax: Axes. Axis into which the plot will be made.

        :param cmap: string. A valid matplolib color map name.

        :param vmin: float. Minimum of the color scaling.

        :param vmax: float. Maximum of the color scaling.

        :param colorbar: bool. Draw a colorbar for the currents.

        :param \**kwargs: dict. All other key word arguments are passed to the mplot3d plot function.

        :return: list of Line3D objects. The line plotted for each coil.

        """

        if ax is None:
            # extra logic to handle OMFIT's auto-figure making when double-clicked in tree
            f = pyplot.gcf()
            if len(f.axes):
                # normal situation in which existing figures should be respected and left alone
                try:
                    f, ax = pyplot.subplots(subplot_kw={'aspect': 'equal', 'projection': '3d'})
                except NotImplementedError:  # python3 matplolib did not have this for some reason
                    pyplot.close(pyplot.gcf())  # will have opened a figure than failed to set the aspect
                    f, ax = pyplot.subplots(subplot_kw={'projection': '3d'})
            else:
                # OMFIT (or the gcf needed to check OMFIT) made a empty figure for us
                ax = pyplot.subplot(111, projection='3d')
                try:
                    ax.set_aspect('equal')
                except NotImplementedError:
                    pass
            # aesthetics
            ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.set_zlabel('z (m)')
            ax.set_axis_off()
        else:
            f = ax.get_figure()

        curs, lines = [], []
        bound = 0
        for k, v in list(self.items()):
            curs.append(v['current'])
            nseg = v['nseg']
            eta = np.linspace(0, 2 * np.pi, nseg + 1, True)
            ms = np.arange(v['nfcoil'] + 1)
            x = np.real(np.sum((v['xc'] - 1j * v['xs'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))
            y = np.real(np.sum((v['yc'] - 1j * v['ys'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))
            z = np.real(np.sum((v['zc'] - 1j * v['zs'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))
            lines.append(ax.plot(x, y, z, label=k, **kwargs)[0])
            bound = max(bound, np.max(np.abs(np.hstack((x, y, z)))))
            if v['coil_symm'] == 1:
                for isym in range(1, bnfp):
                    sx = np.real(x * np.exp(1j * 2 * np.pi * isym / bnfp))
                    sy = np.real(y * np.exp(1j * 2 * np.pi * isym / bnfp))
                    sz = z * 1.0
                    curs.append(v['current'])
                    lines.append(ax.plot(sx, sy, sz, label=f'{k}_symm{isym}', **kwargs)[0])
                    bound = max(bound, np.max(np.abs(np.hstack((sx, sy, sz)))))
        # color by current
        from omfit_classes.utils_plot import set_linearray

        sm = set_linearray(lines, values=curs, vmin=vmin, vmax=vmax, cmap=cmap)
        if colorbar:
            cb = f.colorbar(sm, ax=ax, shrink=0.8)
            cb.set_label('Current (A)')

        # force equal aspect
        bound *= 0.6
        ax.set_xlim(-bound, bound)
        ax.set_ylim(-bound, bound)
        ax.set_zlim(-bound, bound)
        ax.auto_scale_xyz((-bound, bound), (-bound, bound), (-bound, bound))

        return lines

    @dynaLoad
    def _to_OMFITcoils(self, filename, bnfp=1, gpec_format=False):
        """
        Convert to the standard OMFITcoils ascii file format, which stores x,y,z,i points
        by inverse Fourier transforming the FOCUS coil decompositions.

        :param filename: str. Name of new file

        :param bnfp: int. Toroidal mode number of symmetry used if coil_symm is 1

        :param gpec_format: bool. Convert to OMFITGPECcoils ascii formatting instead

        :return: OMFITcoils or OMFITGPECcoils

        """
        coils = SortedDict()
        coils['periods'] = 1
        coils['num_coils'] = len(list(self.keys()))

        for i, (key, coil) in enumerate(list(self.items())):
            nseg = coil['nseg']
            eta = np.linspace(0, 2 * np.pi, nseg + 1, True)
            ms = np.arange(coil['nfcoil'] + 1)
            x = np.real(np.sum((coil['xc'] - 1j * coil['xs'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))
            y = np.real(np.sum((coil['yc'] - 1j * coil['ys'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))
            z = np.real(np.sum((coil['zc'] - 1j * coil['zs'])[:, None] * np.exp(1j * ms[:, None] * eta), 0))

            coils[key] = SortedDict()
            coils[key]['x'] = x
            coils[key]['y'] = y
            coils[key]['z'] = z
            coils[key]['I'] = coil['current'] * 1.0
            coils[key]['group'] = i + 1  # todo: can we get the group from FOCUS coils?
            coils[key]['name'] = key

            # if there is symmetry we need to add appropriately rotated coils
            if coil['coil_symm'] == 1:
                for isym in range(1, bnfp):
                    skey = f'{key}_symm{isym}'
                    coils[skey] = SortedDict()
                    coils[skey]['x'] = np.real(x * np.exp(1j * 2 * np.pi * isym / bnfp))
                    coils[skey]['y'] = np.real(y * np.exp(1j * 2 * np.pi * isym / bnfp))
                    coils[skey]['z'] = z * 1.0
                    coils[skey]['I'] = coil['current'] * 1.0
                    coils[skey]['group'] = i + 1  # todo: Confirm this should match for symm coils
                    coils[skey]['name'] = skey

        # Use the OMFIT class to create an empty file and then update it in memory
        if gpec_format:
            result = OMFITGPECcoils(filename)
        else:
            result = OMFITcoils(filename)
        result.update(coils)

        return result

    def to_OMFITcoils(self, filename, bnfp=1):
        """
        Convert to the standard OMFITcoils ascii file format, which stores x,y,z,i points
        by inverse Fourier transforming the FOCUS coil decompositions.

        :param filename: str. Name of new file

        :param bnfp: int. Toroidal mode number of symmetry used if coil_symm is 1

        :param gpec_format: bool. Convert to OMFITGPECcoils ascii formatting instead

        :return: OMFITcoils or OMFITGPECcoils

        """
        # Use a central functions so we do not duplicate complicated
        # inverse Fourier transformation code
        result = self._to_OMFITcoils(filename, bnfp=bnfp, gpec_format=False)

        return result

    @dynaLoad
    def to_OMFITGPECcoils(self, filename):
        """
        Convert OMFITcoils to the OMFITGPECcoils ascii file format, which writes x,y,z points
        but no current information.

        :param filename: str. Name of new file

        :param bnfp: int. Toroidal mode number of symmetry used if coil_symm is 1

        :return: OMFITGPECcoils

        """
        # Use a central functions so we do not duplicate complicated
        # inverse Fourier transformation code
        result = self._to_OMFITcoils(filename, bnfp=bnfp, gpec_format=True)

        return result


class OMFITGPECcoils(OMFITcoils):
    """
    Class that reads ascii GPEC .dat coil files and converts them to the standard coil file format used by
    the FOCUS and STELLOPT codes.

    NOTE: This will clobber the original ascii formatting!! Do not connect to original file paths!!

    """

    @dynaLoad
    def load(self):
        """Load the file and parse it into a sorted dictionary"""
        # read the ascii file
        try:
            with open(self.filename, 'r') as f:
                line1 = f.readline()
                ncoil, s, nsec, nw = list(map(int, list(map(float, line1.split()))))
            x, y, z = np.genfromtxt(self.filename, skip_header=1).T.reshape(3, ncoil, s, nsec)
            c = nw

            # fill in all the info needed for a coil file
            coils = self
            coils['periods'] = 1
            coils['num_coils'] = ncoil
            nc = 0
            for i in range(ncoil):
                for j in range(s):
                    key = 'coil_{:}'.format(nc)
                    coils[key] = SortedDict()
                    coils[key]['x'] = x[i, j, :]
                    coils[key]['y'] = y[i, j, :]
                    coils[key]['z'] = z[i, j, :]
                    coils[key]['I'] = nw  # should be multiplied by nominal current
                    coils[key]['group'] = i + 1
                    coils[key]['name'] = os.path.split(self.filename)[-1].split('.')[0]
                    nc += 1
            # coils['num_coils'] = nc  # why was this here? already set ncoil above... why was this ncoil + 1?
        except Exception:
            OMFITcoils.load(self)

    @dynaSave
    def save(self):
        """Write the data in standard ascii format"""
        print(self.filename)
        with open(self.filename, 'w') as f:
            # write the defining parameters
            ncoil = self['num_coils']
            s = 1  # have to assume this?
            nsec = len(self[list(self.keys())[-1]]['x'])
            nw = self[list(self.keys())[-1]]['I']
            if nsec < 1e4:
                f.write('{:>5}{:>5}{:>5}{:8.2f}\n'.format(ncoil, s, nsec, nw))  # the first line with periods
            else:
                f.write('{:>5}{:>5} {:}{:8.2f}\n'.format(ncoil, s, nsec, nw))  # the first line with periods
            # write each coil x, y, z
            for key, coil in list(self.items()):
                if not isinstance(coil, dict):
                    continue
                for i in range(len(coil['x'])):
                    f.write('{:13.4e}{:13.4e}{:13.4e}\n'.format(coil['x'][i], coil['y'][i], coil['z'][i]))
        return

    @dynaLoad
    def to_OMFITcoils(self, filename):
        """
        Convert to the standard OMFITcoils ascii file format, which stores x,y,z,i points.

        :param filename: str. Name of new file

        :return: OMFITcoils

        """

        # Use the OMFIT class to create an empty file and then update it in memory
        result = OMFITcoils(filename)
        result.update(self)

        return result



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.fluxSurface import rz_miller
from omfit_classes.omfit_matlab import OMFITmatlab
from omfit_classes.utils_math import atomic_element
from omas import ODS
import numpy as np

__all__ = ['OMFITcotsim']


class OMFITcotsim(OMFITmatlab):
    def to_omas(self, ods=None, time_index=None):
        if ods is None:
            ods = ODS()
        time = self['outputs']['scalars']['t']
        if time_index is None:
            time_index = range(len(time))

        # shouldn't all these be time dependent
        r, _ = np.meshgrid(self['outputs']['config']['model']['params']['r']['value'], time)
        R0 = self['outputs']['config']['model']['params']['R0']['value']
        B0 = self['outputs']['config']['model']['params']['B0']['value'] * time
        kappa, _ = np.meshgrid(self['outputs']['config']['model']['params']['kappa_prof']['value'], time)
        delta, _ = np.meshgrid(self['outputs']['config']['model']['params']['delta_prof']['value'], time)
        sshaf, _ = np.meshgrid(self['outputs']['config']['model']['params']['sshaf_prof']['value'] * 0.01, time)
        zeta = delta * 0.0
        zmag = delta * 0.0
        rmag = sshaf + R0

        for ktime_index, time_index in enumerate(tolist(time_index)):
            ods[f'core_profiles.profiles_1d.{ktime_index}.electrons.density_thermal'] = ne = self['outputs']['profiles']['ne'][
                time_index, :
            ]
            ods[f'core_profiles.profiles_1d.{ktime_index}.electrons.temperature'] = te = (
                self['outputs']['profiles']['te'][time_index, :] * 1e3
            )
            ods[f'core_profiles.profiles_1d.{ktime_index}.grid.rho_tor_norm'] = rho = self['outputs']['profiles']['rho_norm']
            for kion, ion in enumerate(['D', 'C']):
                ion_info = list(atomic_element(symbol=ion).values())[0]
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.density_thermal'] = self['outputs']['profiles']['n' + ion][
                    time_index, :
                ]
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.element.0.a'] = ion_info['A']
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.element.0.atoms_n'] = 1
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.element.0.z_n'] = ion_info['Z']
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.label'] = ion
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.multiple_states_flag'] = 0
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.temperature'] = ti = (
                    self['outputs']['profiles']['ti'][time_index, :] * 1e3
                )
                ods[f'core_profiles.profiles_1d.{ktime_index}.ion.{kion}.z_ion'] = 1.0
            ods[f'core_profiles.profiles_1d.{ktime_index}.rotation_frequency_tor_sonic'] = self['outputs']['profiles']['omega'][
                time_index, :
            ]
            ods[f'core_profiles.profiles_1d.{ktime_index}.zeff'] = rho * 0.0 + 1.0
            ods.set_time_array('core_profiles.vacuum_toroidal_field.b0', ktime_index, B0[time_index])
            ods[f'core_profiles.vacuum_toroidal_field.r0'] = R0
            ods.set_time_array('core_profiles.time', ktime_index, time[time_index])

            r_boundary, z_boundary = rz_miller(
                r[time_index, -1],
                rmag[time_index, -1],
                kappa=kappa[time_index, -1],
                delta=delta[time_index, -1],
                zeta=zeta[time_index, -1],
                zmag=zmag[time_index, -1],
            )
            ods[f'equilibrium.time_slice.{ktime_index}.boundary.outline.r'] = r_boundary
            ods[f'equilibrium.time_slice.{ktime_index}.boundary.outline.z'] = z_boundary
            ods[f'equilibrium.time_slice.{ktime_index}.global_quantities.ip'] = self['outputs']['scalars']['ip'][time_index] * 1e6
            ods[f'equilibrium.time_slice.{ktime_index}.global_quantities.magnetic_axis.b_field_tor'] = (
                B0[time_index] * R0 / rmag[time_index, 0]
            )
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.centroid.r_max'] = rmag[time_index, :] + r[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.centroid.r_min'] = rmag[time_index, :] - r[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.centroid.z'] = zmag[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.elongation'] = kappa[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.psi'] = self['outputs']['profiles']['psi'][time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.q'] = self['outputs']['profiles']['q'][time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.rho_tor_norm'] = rho
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.squareness_lower_inner'] = zeta[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.squareness_lower_outer'] = zeta[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.squareness_upper_inner'] = zeta[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.squareness_upper_outer'] = zeta[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.triangularity_lower'] = delta[time_index, :]
            ods[f'equilibrium.time_slice.{ktime_index}.profiles_1d.triangularity_upper'] = delta[time_index, :]
            ods.set_time_array('equilibrium.vacuum_toroidal_field.b0', ktime_index, B0[time_index])
            ods[f'equilibrium.vacuum_toroidal_field.r0'] = R0
            ods.set_time_array('equilibrium.time', ktime_index, time[time_index])

        # calculate Zeff
        ods.physics_core_profiles_zeff()
        return ods



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import numpy as np
from omfit_classes.omfit_path import OMFITpath
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.sortedDict import OMFITdataset
from omfit_classes.omfit_dir import OMFITdir
from omfit_classes.omfit_gacode import OMFITgacode
from omfit_classes.omfit_asciitable import OMFITasciitable

__all__ = [
    'OMFITtglfEPinput',
    'OMFITalphaInput',
    'OMFITtglf_eig_spectrum',
    'OMFITtglf_wavefunction',
    'OMFITtglf_flux_spectrum',
    'OMFITtglf_nete_crossphase_spectrum',
    'OMFITtglf_potential_spectrum',
    'OMFITtglf_fluct_spectrum',
    'OMFITtglf_intensity_spectrum',
    'OMFITtglf',
    'OMFITtglf_nsts_crossphase_spectrum',
    'sum_ky_spectrum',
]


class OMFITtglfEPinput(OMFITascii, SortedDict):
    '''
    Class used to read/write TGLFEP input files
    '''

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read()

        widths = []
        for line in lines.split('\n'):
            if not len(line.strip()):
                continue
            try:
                if len(widths):
                    self['WIDTHS'] = widths
                value, key = line.split()
            except ValueError:  # not enough values to unpack
                widths.append(eval(line))
            else:
                if value == '.false.':
                    value = False
                elif value == '.true.':
                    value = True
                else:
                    value = eval(value)
                self[key] = value

    @dynaSave
    def save(self):
        txt = []
        for key, value in self.items():
            if key == 'WIDTHS':
                continue
            if value is False:
                value = '.false.'
            elif value is True:
                value = '.true.'
            txt.append(f'{value} {key}')
            if key == 'WIDTH_IN_FLAG' and 'WIDTHS' in self:
                txt.append('\n'.join(map(str, self['WIDTHS'])))
        with open(self.filename, 'w') as f:
            f.write('\n'.join(txt))


class OMFITalphaInput(OMFITascii, SortedDict):
    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read().strip().split('\n')
        comment = lines[0]
        data = np.array(list(map(eval, lines[1:])))
        self['comment'] = comment
        self['data'] = data


def get_ky_spectrum(filename):
    with open(filename, 'r') as f:
        content = f.readlines()
    content = ''.join(content[2:]).split()
    ky = np.array(content, dtype=float)
    return ky


class OMFITtglf_eig_spectrum(OMFITdataset, OMFITascii):
    def __init__(self, filename, ky_file, nmodes, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.nmodes = self.OMFITproperties['nmodes'] = nmodes

    @dynaLoad
    def load(self):
        nmodes = self.nmodes
        with open(self.filename, 'r') as f:
            content = f.readlines()
        content = ''.join(content[2:]).split()
        ky = get_ky_spectrum(self.ky_file)
        gamma = []
        freq = []
        for k in range(self.nmodes):
            gamma.append(np.array(content[2 * k :: nmodes * 2], dtype=float))
            freq.append(np.array(content[2 * k + 1 :: nmodes * 2], dtype=float))
        gamma = np.array(gamma)
        freq = np.array(freq)
        gamma = DataArray(gamma, dims=('mode_num', 'ky'), coords={'ky': ky, 'mode_num': np.arange(nmodes) + 1})
        freq = DataArray(freq, dims=('mode_num', 'ky'), coords={'ky': ky, 'mode_num': np.arange(nmodes) + 1})
        self.update({'gamma': gamma, 'freq': freq})

    def plot(self, axs=None):
        '''
        Plot the eigenvalue spectrum as growth rate and frequency

        :param axs: A length 2 sequence of matplotlib axes (growth rate, frequency)
        '''
        if axs is None:
            fig, axs = pyplot.subplots(2, 1, num=pyplot.gcf().number, squeeze=True, sharex=True)
        else:
            fig = axs[0].get_figure()
        for k in range(self.nmodes):
            axs[0].plot(self['ky'], self['gamma(%d)' % (k + 1)], label='Growth rate $\\gamma_{%d}$' % (k + 1))
            axs[1].plot(self['ky'], self['freq(%d)' % (k + 1)], label='Frequency $\\omega_{%d}$' % (k + 1))
        axs[0].set_ylabel('Growth rate $\\gamma$')
        with warnings.catch_warnings():
            # We don't care if matplotlib has an underflow while drawing
            warnings.filterwarnings('ignore', category=RuntimeWarning, message='underflow*')
            axs[0].set_yscale('symlog', linthresh=0.01)
            axs[1].set_yscale('symlog', linthresh=0.01)
            axs[1].set_ylabel('Frequency $\\omega$')
            axs[1].axhline(0, color='gray', ls='--')
            axs[1].set_xlabel('$k_\\perp \\rho$')
            axs[0].set_xscale('log')
            fig.canvas.draw()

    @dynaLoad
    def __getitem__(self, key):

        if key in self:
            return OMFITdataset.__getitem__(self, key)

        if key.startswith('gamma') and '(' in key:
            ind = int(key.split('(')[1].split(')')[0])
            return self['gamma'].sel(mode_num=ind).values

        if key.startswith('freq') and '(' in key:
            ind = int(key.split('(')[1].split(')')[0])
            return self['freq'].sel(mode_num=ind).values

        raise KeyError(key)


class OMFITtglf_wavefunction(OMFITascii, SortedDict):
    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        nmodes, nfields, npoints = list(map(int, lines[0].split()))
        self.nmodes = nmodes
        self.nfields = nfields
        keys = lines[1].split()
        self.headers = keys[1:]
        content = ''.join(lines[2:]).split()
        self['theta'] = np.array(content[0 :: nmodes * nfields * 2 + 1], dtype=float)
        for ik, k in enumerate(keys[1:]):
            for n in range(nmodes):
                self['%s(%d)' % (k, n + 1)] = np.array(content[2 * n * nfields + 1 + ik :: nmodes * nfields * 2 + 1], dtype=float)

    def plot(self):
        def pi_mult_formatter(x, pos):
            if x == 0:
                return '0'
            if x == np.pi:
                return r'$\pi$'
            if x == -np.pi:
                return r'$-\pi$'
            return r'$%d\pi$' % (int(round(x / np.pi, 0)))

        fig, axs = pyplot.subplots(self.nfields, self.nmodes, num=pyplot.gcf().number, sharex=True, sharey='row', squeeze=False)
        for n in range(self.nmodes):
            for ki, k in enumerate(self.headers):
                axs[int(ki // 2)][n].plot(self['theta'], self['%s(%d)' % (k, n + 1)], ls=['-', '--'][ki % 2])
                if n == 0:
                    axs[int(ki // 2)][n].set_ylabel(k.split('(')[1].split(')')[0])
                if ki == len(self.headers) - 1:
                    axs[int(ki // 2)][n].set_xlabel(r'$\theta$')
                    axs[int(ki // 2)][n].xaxis.set_major_locator(MultipleLocator(np.pi))
                    axs[int(ki // 2)][n].xaxis.set_major_formatter(FuncFormatter(pi_mult_formatter))
        fig.canvas.draw()


class OMFITtglf_flux_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the out.tglf.sum_flux_spectrum file and provide a convenient means for
    plotting it.

    :param filename: Path to the out.tglf.sum_flux_spectrum file

    :param n_species: Number of species

    :param n_fields: Number of fields
    '''

    def __init__(self, filename, ky_file, n_species, n_fields, spec_labels, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.n_species = self.OMFITproperties['n_species'] = n_species
        self.n_fields = self.OMFITproperties['n_fields'] = n_fields
        self.spec_labels = self.OMFITproperties['spec_labels'] = spec_labels

    @dynaLoad
    def load(self):
        ns = self.n_species
        nf = self.n_fields
        with open(self.filename) as f:
            content = f.read()
        tmpdict = SortedDict()
        ky = get_ky_spectrum(self.ky_file)
        tmpdict['ky'] = np.array(ky, dtype=float)
        for s in range(ns):
            for f in range(nf):
                ind1 = content.find('species =')
                ind2 = content.find('species =', ind1 + 10)
                data = content[ind1:ind2]
                first_two = data.splitlines()[:2]
                spec = int(first_two[0].split()[2])
                field = int(first_two[0].split()[5])
                spec_label = self.spec_labels[s]
                labels = [k.strip().replace('flux', '').strip() for k in first_two[1].split(',')]
                keys = [k + '_field_%d_spec_%s' % (field, spec_label) for k in labels]
                labels = labels[:]
                data = ''.join(data.splitlines()[2:]).split()
                for ki, k in enumerate(keys):
                    tmpdict[k] = np.array(data[ki :: len(keys)], dtype=float)
                content = content[content.find('species =', 10) :]
        ky = tmpdict['ky']
        tmparray = {}
        for label in labels:
            tmparray['_'.join(label.split())] = DataArray(
                np.zeros((ns, nf, len(ky))),
                dims=('species', 'field', 'ky'),
                coords={'species': self.spec_labels, 'field': ['phi', 'B_perp', 'B_parallel'][: self.n_fields], 'ky': ky},
            )
        for k, v in list(tmpdict.items()):
            if k == 'ky':
                continue
            quant, field = k.split('_field_', 2)
            quant = '_'.join(quant.split())
            field, spec_label = field.split('_spec_')
            spec_ind = self.spec_labels.index(spec_label)
            field_ind = int(field) - 1
            tmparray[quant][spec_ind, field_ind, :] = v
        self.update(tmparray)
        self._initialized = False
        self.labels = labels
        self._initialized = True

    @dynaLoad
    def plot(self, fn=None):
        '''
        Plot the flux spectra

        :param fn: A FigureNotebook instance
        '''
        ns = self.n_species
        nf = self.n_fields
        if fn is None:
            from omfit_plot import FigureNotebook

            tabbed = FigureNotebook(nfig=0, name='TGLF Flux Spectra')
        else:
            tabbed = fn
        for k in self.labels:
            fig = tabbed.add_figure(label=k)
            fig.suptitle(k)
            for s in range(ns):
                for f in range(nf):
                    if s == 0 and f == 0:
                        ax0 = ax = axr = fig.add_subplot(ns, nf, s * nf + f + 1)  # OKadd
                    elif f == 0:
                        ax = axr = fig.add_subplot(ns, nf, s * nf + f + 1, sharex=ax0)  # OKadd
                    else:
                        ax = fig.add_subplot(ns, nf, s * nf + f + 1, sharex=ax0, sharey=axr)  # OKadd
                    if f == 0:
                        ax.set_ylabel('Species: %s' % (self.spec_labels[s]))
                    if s == 0:
                        ax.set_title([r'$\phi$', r'$B_{\perp}$', r'$B_{\parallel}$'][f])
                    ax.plot(self['ky'], self[k + '_field_%d_spec_%s' % (f + 1, self.spec_labels[s])])
                    ax.set_xscale('log')
                    ax.axis('tight')
            autofmt_sharexy(fig=fig)
            fig.canvas.draw()

    @dynaLoad
    def __getitem__(self, key):
        if key in self:
            return OMFITdataset.__getitem__(self, key)
        try:
            quant, field = key.split('_field_', 2)
            quant = '_'.join(quant.split())
            field, spec_label = field.split('_spec_')
            spec_ind = self.spec_labels.index(spec_label)
            field_ind = int(field) - 1
            return self[quant].isel(field=field_ind, species=spec_ind)
        except Exception as _excp:
            print(_excp)
            raise KeyError(k)


class OMFITtglf_nete_crossphase_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the out.tglf.nete_crossphase_spectrum file and provide a convenient means for
    plotting it.

    :param filename: Path to the out.tglf.nete_crossphase_spectrum file

    :param nmodes: Number of modes computed by TGLF and used in the
    '''

    def __init__(self, filename, ky_file, nmodes, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.nmodes = self.OMFITproperties['nmodes'] = nmodes
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file

    @dynaLoad
    def load(self):
        nmodes = self.nmodes
        with open(self.filename, 'r') as f:
            content = f.readlines()
        content = ''.join(content[2:]).split()
        ky = get_ky_spectrum(self.ky_file)
        tmparray = []
        for k in range(self.nmodes):
            tmparray.append(np.array(content[k::nmodes], dtype=float))
        self.update({'nete': DataArray(np.array(tmparray), dims=('mode_num', 'ky'), coords={'mode_num': np.arange(nmodes) + 1, 'ky': ky})})

    def plot(self, ax=None):
        '''
        Plot the nete crossphase spectrum

        :param ax: A matplotlib axes instance
        '''
        if ax is None:
            fig, ax = pyplot.subplots(1, 1, squeeze=True)
        ky = self['ky']
        nmodes = self.nmodes
        for k in range(nmodes):
            ax.plot(ky, self['nete_crossphase_spectrum(%d)' % (k + 1)] / np.pi, label='Mode %d' % (k + 1))
        ax.legend(loc='best')
        ax.set_xscale('log')
        ax.set_xlabel(r'$k_y$')
        ax.set_ylabel(r'$\delta n_e$ $\delta T_e$ crossphase')
        ax.set_title(r'$\Theta_{\delta n_e\delta T_e}$')
        ax.axis('tight')
        ax.set_yticks(np.linspace(-np.pi / 2.0, np.pi / 2.0, 5), [r'$-\pi/2$', r'$\pi/4$', '0', r'$\pi/4$', r'$\pi/2$'])
        fmt = pyplot.ScalarFormatter()
        fmt.set_scientific(False)
        ax.xaxis.set_major_formatter(fmt)

    @dynaLoad
    def __getitem__(self, key):
        if key in self:
            return OMFITdataset.__getitem__(self, key)
        if 'nete_crossphase_spectrum(' in key:
            ind = int(key.split('(')[1].split(')')[0])
            return self['nete'].sel(mode_num=ind)
        raise KeyError(key)


class OMFITtglf_nsts_crossphase_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the out.tglf.nsts_crossphase_spectrum file and provide a convenient means for
    plotting it.

    :param filename: Path to the out.tglf.nsts_crossphase_spectrum file

    :param nmodes: Number of modes computed by TGLF and used in the
    '''

    def __init__(self, filename, ky_file, nmodes, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.nmodes = self.OMFITproperties['nmodes'] = nmodes

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            content = f.readlines()  # Remove the main header
        header = content.pop(0)
        nspec = int(header.strip().split()[-2])
        nmodes = self.nmodes
        for ci, c in enumerate(content):
            if ci == 0:
                continue
            if c.strip().startswith('species'):
                nk = ci - 2  # Two header lines
                break
        nsts = np.zeros((nspec, nmodes, nk), dtype=float)
        for si in range(nspec):
            block = ' '.join(content[(2 + nk) * si + 2 : (2 + nk) * (si + 1)]).split()
            if si == 0:
                ky = get_ky_spectrum(self.ky_file)
            for mi in range(nmodes):
                nsts[si, mi, :] = np.array(block[mi::nmodes], dtype=float)
        self.update(
            {
                'nsts': DataArray(
                    np.array(nsts),
                    dims=('spec', 'mode_num', 'ky'),
                    coords={'spec': np.arange(nspec) + 1, 'mode_num': np.arange(nmodes) + 1, 'ky': ky},
                )
            }
        )

    def plot(self, axs=None):
        '''
        Plot the nsts crossphase spectrum

        :param axs: A sequence of matplotlib axes instances of length len(self['spec'])
        '''
        ky = self['ky']
        nmodes = self.nmodes
        nspec = len(self['spec'])
        if axs is None or len(np.atleast1d(axs)) != nspec:
            fig, axs = pyplot.subplots(nspec, 1, squeeze=True, sharex=True)
        fmt = pyplot.ScalarFormatter()
        fmt.set_scientific(False)
        for si, spec in enumerate(self['spec']):
            ax = axs[si]
            ax.set_ylabel(r'$\delta n_s$ $\delta T_s$ crossphase')
            ax.xaxis.set_major_formatter(fmt)
            ax.set_yticks(np.linspace(-np.pi / 2.0, np.pi / 2.0, 5), [r'$-\pi/2$', r'$\pi/4$', '0', r'$\pi/4$', r'$\pi/2$'])
            for mi in range(nmodes):
                ax.plot(ky, self['nsts'].isel(spec=si, mode_num=mi) / np.pi, label='Mode %d' % (mi + 1))
        ax.legend(loc='best')
        ax.set_xscale('log')
        ax.set_xlabel(r'$k_y$')
        axs[0].set_title(r'$\Theta_{\delta n_s\delta T_s}$')
        ax.axis('tight')


class OMFITtglf_potential_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the potential fluctuation spectrum in out.tglf.potential_spectrum and
    provide a convenient means for plotting it.

    :param filename: Path to the out.tglf.potential_spectrum file
    '''

    def __init__(self, filename, ky_file, nmodes, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.nmodes = self.OMFITproperties['nmodes'] = nmodes

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        nmodes = self.nmodes
        ky = get_ky_spectrum(self.ky_file)
        description = lines[0]
        columns = [x.strip() for x in lines[1].split(',')]
        nc = len(columns)
        content = ''.join(lines[6:]).split()
        tmpdict = {}
        tmpdict['ky'] = np.array(ky, dtype=float)
        for ik, k in enumerate(columns):
            tmp = []
            for nm in range(nmodes):
                tmp.append(np.array(content[ik + nm * nc :: nmodes * nc], dtype=float))
            tmpdict[k] = tmp
        for k, v in list(tmpdict.items()):
            if k == 'ky':
                continue
            self[k] = xarray.DataArray(v, dims=('mode_num', 'ky'), coords={'ky': tmpdict['ky'], 'mode_num': np.arange(nmodes) + 1})

    @dynaLoad
    def plot(self, fn=None):
        '''
        Plot the fields

        :param fn: A FigureNotebook instance
        '''
        from omfit_plot import FigureNotebook

        if fn is None:
            fn = FigureNotebook('Field spectra')
        nmodes = self.nmodes
        fmt = pyplot.ScalarFormatter()
        fmt.set_scientific(False)
        for k in self.data_vars:
            for nm in range(nmodes):
                if not np.any(self[k] != 0):
                    continue
                fig, ax = fn.subplots(label=k + ', mode = ' + str(nm + 1), squeeze=True)
                ax.plot(self['ky'], self[k][nm])
                ax.set_xscale('log')
                ax.axis('tight')
                ax.xaxis.set_major_formatter(fmt)
                ax.set_xlabel('$k_y$')


class OMFITtglf_fluct_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the {density,temperature} fluctuation spectrum in
    out.tglf.{density,temperature}_spectrum  and provide a convenient means for
    plotting it.

    :param filename: Path to the out.tglf.{density,temperature}_spectrum file

    :param ns: Number of species

    :param label: Type of fluctuations ('density' or 'temperature')
    '''

    pretty_names = {'density': r'\delta n ', 'temperature': r'\delta T '}

    def __init__(self, filename, ky_file, ns, label, spec_labels, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.ns = self.OMFITproperties['ns'] = ns
        self.label = self.OMFITproperties['label'] = label
        self.spec_labels = self.OMFITproperties['spec_labels'] = spec_labels

    @dynaLoad
    def load(self):
        ns = self.ns
        with open(self.filename, 'r') as f:
            content = f.readlines()
        content = ''.join(content[2:]).split()
        ky = get_ky_spectrum(self.ky_file)
        tmplist = []
        for s in range(ns):
            tmplist.append(np.array(content[s::ns], dtype=float))
        self[self.label] = xarray.DataArray(np.array(tmplist), dims=('species', 'ky'), coords={'species': self.spec_labels, 'ky': ky})

    @dynaLoad
    def plot(self, axs=None):
        '''
        Plot the fluctuation spectrum

        :param axs: A list of matplotlib axes of length self.ns
        '''
        from matplotlib import pyplot
        from matplotlib.ticker import ScalarFormatter

        if axs is None:
            fig, axs = pyplot.subplots(self.ns, 1, sharex=True, sharey=True)
        else:
            if len(axs) != self.ns:
                raise ValueError('Must pass length %s list of axes' % self.ns)
            fig = axs[0].get_figure()
        ns = self.ns
        lab_base = self.pretty_names[self.label]
        for s in range(ns):
            ax = axs[s]
            ax.plot(self['ky'], self[self.label + '_spec_%d' % (s + 1)], label=r'$%s_{%s}$' % (lab_base, self.spec_labels[s]))
            ax.legend()
            ax.set_xscale('log')
            ax.axis('tight')
            fmt = pyplot.ScalarFormatter()
            fmt.set_scientific(False)
            ax.xaxis.set_major_formatter(fmt)
        ax.set_xlabel(r'$k_y$')
        try:
            autofmt_sharexy(fig=fig)
        except NameError:
            pass

    def __getitem__(self, key):
        if key in self:
            return OMFITdataset.__getitem__(self, key)
        if key.startswith(self.label + '_spec_'):
            spec_ind = int(key.split('_')[-1]) - 1
            return self[self.label].isel(species=spec_ind)
        raise KeyError(key)


class OMFITtglf_intensity_spectrum(OMFITdataset, OMFITascii):
    '''
    Parse the intensity fluctuation spectrum in
    out.tglf.{density,temperature}_spectrum  and provide a convenient means for
    plotting it.

    :param filename: Path to the out.tglf.{density,temperature}_spectrum file

    :param ns: Number of species

    :param label: Type of fluctuations ('density' or 'temperature')
    '''

    def __init__(self, filename, ky_file, nmodes, ns, spec_labels, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True
        OMFITdataset.__init__(self)
        self.ky_file = self.OMFITproperties['ky_file'] = ky_file
        self.nmodes = self.OMFITproperties['nmodes'] = nmodes
        self.ns = self.OMFITproperties['ns'] = ns
        self.spec_labels = self.OMFITproperties['spec_labels'] = spec_labels

    @dynaLoad
    def load(self):
        ns = self.ns
        nmodes = self.nmodes
        ky = get_ky_spectrum(self.ky_file)
        nky = len(ky)
        with open(self.filename, 'r') as f:
            dens, temp, vpara, enpara = np.loadtxt(f, skiprows=4, unpack=True)

        density = dens.reshape(ns, nky, nmodes)
        self['density'] = xarray.DataArray(
            density, dims=('ns', 'ky', 'nmodes'), coords={'ns': np.arange(ns) + 1, 'ky': ky, 'nmodes': np.arange(nmodes) + 1}
        )

        temperature = temp.reshape(ns, nky, nmodes)
        self['temperature'] = xarray.DataArray(
            temperature, dims=('ns', 'ky', 'nmodes'), coords={'ns': np.arange(ns) + 1, 'ky': ky, 'nmodes': np.arange(nmodes) + 1}
        )

        parallel_vel = vpara.reshape(ns, nky, nmodes)
        self['parallel_velocity'] = xarray.DataArray(
            parallel_vel, dims=('ns', 'ky', 'nmodes'), coords={'ns': np.arange(ns) + 1, 'ky': ky, 'nmodes': np.arange(nmodes) + 1}
        )

        parallel_energy = enpara.reshape(ns, nky, nmodes)
        self['parallel_energy'] = xarray.DataArray(
            parallel_energy, dims=('ns', 'ky', 'nmodes'), coords={'ns': np.arange(ns) + 1, 'ky': ky, 'nmodes': np.arange(nmodes) + 1}
        )


class OMFITtglf(OMFITdir):
    '''
    The purpose of this class is to be able to store all results from a given
    TGLF run in its native format, but parsing the important parts into the tree

    :param filename: Path to TGLF run
    '''

    def __init__(self, filename, **kw):
        printd('Calling OMFITtglf init', topic='OMFITtglf')
        OMFITdir.__init__(self, filename, **kw)
        printd('print self.keys():', list(self.keys()), topic='OMFITtglf')

        # remove 'OMFIT_run_command.sh'
        if 'OMFIT_run_command.sh' in self:
            del self['OMFIT_run_command.sh']

        # convert OMFITpath to OMFITascii
        for item in list(self.keys()):
            if isinstance(self[item], OMFITpath):
                self[item].__class__ = OMFITascii

        # rename input.tglf.gen to input.tglf
        if 'input.tglf.gen' in self:
            input_tglf = self['input.tglf'] = OMFITgacode(self['input.tglf.gen'].filename, noCopyToCWD=True)
            del self['input.tglf.gen']

        ns = self['input.tglf']['NS']
        from omfit_classes.utils_math import element_symbol

        self.spec_labels = [element_symbol(self['input.tglf']['ZS_%d' % x]).lower() for x in range(1, ns + 1)]

        # Note: as part of the parsing process, some entries under SELF are deleted.
        #      however we are not deleting the files these entries referred to,
        #      and because OMFITtglf inherits from OMFITdir these files will be carried
        #      through even if they are not visible in the OMFIT tree
        if np.any([str(k).startswith('out') for k in self]):
            self._parse_version()
            self._parse_prec()
            if not input_tglf['USE_TRANSPORT_MODEL']:  # linear run
                self['wavefunction'] = OMFITtglf_wavefunction(self['out.tglf.wavefunction'].filename, noCopyToCWD=True)
                del self['out.tglf.wavefunction']
            else:
                if 'out.tglf.eigenvalue_spectrum' in self:
                    self['eigenvalue_spectrum'] = OMFITtglf_eig_spectrum(
                        self['out.tglf.eigenvalue_spectrum'].filename,
                        self['out.tglf.ky_spectrum'].filename,
                        nmodes=input_tglf['NMODES'],
                        noCopyToCWD=True,
                    )
                    del self['out.tglf.eigenvalue_spectrum']
                try:
                    self._parse_potential_spectrum()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_flucs_spectra()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_flux_spectrum()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_nete_crossphase_spectrum()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_nsts_crossphase_spectrum()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_gbflux()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_run()
                except Exception as _excp:
                    printe(_excp)
                try:
                    self._parse_intensity_spectrum()
                except Exception as _excp:
                    printe(_excp)

    def plot(self):
        for k in self:
            if hasattr(self[k], 'plot'):
                self[k].plot()

    def __delitem__(self, key):
        '''
        Deleting an item only deletes it from the tree, not from the underlying
        directory (which ``OMFITdir`` would do)
        '''
        super(OMFITdir, self).__delitem__(key)

    def _parse_flucs_spectra(self):
        '''
        Parse the fluctuation spectra in out.tglf.{density,temperature}_spectrum
        '''
        ns = self['input.tglf']['NS']
        result = SortedDict()
        for field in ['density', 'temperature']:
            if '%s_spectrum' % field in self:
                continue
            fn = 'out.tglf.%s_spectrum' % field
            if fn not in self:
                printe('No %s file to parse' % fn)
                continue
            self['%s_spectrum' % field] = OMFITtglf_fluct_spectrum(
                self[fn].filename, self['out.tglf.ky_spectrum'].filename, ns, field, self.spec_labels, noCopyToCWD=True
            )
            del self[fn]

    def _parse_potential_spectrum(self):
        '''
        Parse the potential fluctuation spectrum in out.tglf.potential_spectrum
        '''
        nm = self['input.tglf']['NMODES']
        if 'potential_spectrum' in self:
            return
        for fn in ['out.tglf.potential_spectrum', 'out.tglf.field_spectrum']:
            if fn not in self and fn == 'out.tglf.potential_spectrum':
                continue
            elif fn not in self and fn == 'out.tglf.field_spectrum':
                printe('No %s file to parse' % fn)
                return
        self['potential_spectrum'] = OMFITtglf_potential_spectrum(
            self[fn].filename, self['out.tglf.ky_spectrum'].filename, nmodes=nm, noCopyToCWD=True
        )
        del self[fn]

    def _parse_version(self):
        '''
        Parse the version information in out.tglf.version
        '''
        with open(self['out.tglf.version'].filename, 'r') as f:
            content = f.readlines()
        self['GACODE_VERSION'], self['GACODE_PLATFORM'], self['TIMESTAMP'] = [x.strip() for x in content[:3]]
        del self['out.tglf.version']

    def _parse_prec(self):
        '''
        Parse the out.tglf.prec file (a single number)
        '''
        self['regression_prec'] = None
        if 'out.tglf.prec' in self:
            with open(self['out.tglf.prec'].filename, 'r') as f:
                content = f.read()
            self['regression_prec'] = float(content.strip())
            del self['out.tglf.prec']

    def _parse_flux_spectrum(self):
        '''
        Parse the flux spectrum from out.tglf.sum_flux_spectrum
        '''
        ns = self['input.tglf']['NS']
        nf = 1 + self['input.tglf']['USE_BPAR'] + self['input.tglf']['USE_BPER']
        spec_labels = []
        spec_labels_counter = {}
        for item in range(1, ns + 1):
            A = self['input.tglf']['ZS_%d' % item]
            Z = self['input.tglf']['MASS_%d' % item] * 2
            try:
                element = str(element_symbol(A, Z).lower())
            except ValueError:
                element = 'lump'
            spec_labels_counter.setdefault(element, []).append(element)
            if len(spec_labels_counter[element]) == 1:
                spec_labels.append(element)
            else:
                spec_labels.append(element + '%d' % len(spec_labels_counter[element]))
        self['sum_flux_spectrum'] = OMFITtglf_flux_spectrum(
            self['out.tglf.sum_flux_spectrum'].filename, self['out.tglf.ky_spectrum'].filename, ns, nf, spec_labels, noCopyToCWD=True
        )
        del self['out.tglf.sum_flux_spectrum']

    def _parse_nete_crossphase_spectrum(self):
        '''
        Parse the cross phase spectrum from out.tglf.nete_crossphase_spectrum
        '''
        input_tglf = self['input.tglf']
        self['nete_crossphase_spectrum'] = OMFITtglf_nete_crossphase_spectrum(
            self['out.tglf.nete_crossphase_spectrum'].filename,
            self['out.tglf.ky_spectrum'].filename,
            nmodes=input_tglf['NMODES'],
            noCopyToCWD=True,
        )
        del self['out.tglf.nete_crossphase_spectrum']

    def _parse_nsts_crossphase_spectrum(self):
        '''
        Parse the cross phase spectrum from out.tglf.nsts_crossphase_spectrum
        '''
        input_tglf = self['input.tglf']
        self['nsts_crossphase_spectrum'] = OMFITtglf_nsts_crossphase_spectrum(
            self['out.tglf.nsts_crossphase_spectrum'].filename,
            self['out.tglf.ky_spectrum'].filename,
            nmodes=input_tglf['NMODES'],
            noCopyToCWD=True,
        )
        del self['out.tglf.nsts_crossphase_spectrum']

    def _parse_gbflux(self):
        '''
        Parse the flux from the out.tglf.gbflux file
        '''
        try:
            with open(self['out.tglf.gbflux'].filename) as f:
                content = f.read()
            tmp = list(map(float, content.split()))
            if 'std.tglf.gbflux' in self:
                with open(self['std.tglf.gbflux'].filename) as f:
                    content = f.read()
                tmp = uarray(tmp, list(map(float, content.split())))
            tmp = tolist(np.reshape(tmp, (4, -1)))
            species = ['elec'] + ['ion' + str(k) for k in range(1, len(tmp[0]))]
            tmp.insert(0, species)
            printd('Right before out.tglf.run creation', topic='OMFITtglf')
            result = SortedDict()
            result['header'] = ''
            result['columns'] = ['.', 'Gam/Gam_GB', 'Q/Q_GB', 'Pi/Pi_GB', 'S/S_GB']
            result['data'] = torecarray(tmp, result['columns'])
            if 'out.tglf.gbflux' in self:
                del self['out.tglf.gbflux']
            if 'std.tglf.gbflux' in self:
                del self['std.tglf.gbflux']
        except ValueError:
            result = OMFITasciitable(self.filename + '/out.tglf.gbflux', noCopyToCWD=True)
        self['gbflux'] = result

    def _parse_run(self):
        '''
        Parse the flux from the out.tglf.run file
        data should be same as gbflux file, but also includes local MHD stability
        '''
        with open(self['out.tglf.run'].filename) as f:
            lines = f.readlines()
        results = SortedDict()
        lines = [line[:-2] for line in lines]  # Remove '\n' from end of lines
        lines = [line.split() for line in lines]
        for line in lines:
            if 'D(R)' in line:
                for i, item in enumerate(line):
                    try:
                        results[line[i - 2]] = float(item)
                    except Exception:
                        pass
            elif 'Gam' in line[0]:
                results['.'] = line

            elif 'elec' in line[0] or 'ion' in line[0]:
                line_name = line[0]
                line_split = [float(l) for l in line[1:]]
                results[line_name] = line_split

        self['run'] = results
        del self['out.tglf.run']

    def _parse_intensity_spectrum(self):
        input_tglf = self['input.tglf']
        self['intensity_spectrum'] = OMFITtglf_intensity_spectrum(
            self.filename + '/out.tglf.intensity_spectrum',
            self['out.tglf.ky_spectrum'].filename,
            input_tglf['NMODES'],
            input_tglf['NS'],
            self.spec_labels,
        )
        del self['out.tglf.intensity_spectrum']

    def saturation_rule(self, saturation_rule_name):
        ky_spect = np.asarray(self['eigenvalue_spectrum']['ky'])
        gammas = np.asarray(self['eigenvalue_spectrum']['gamma']).T
        potential = self['potential_spectrum']['potential'].T.values

        with open(self['out.tglf.QL_flux_spectrum'].filename, 'r') as f:
            lines = f.readlines()
        nky, nm, ns, nfield, ntype = list(map(int, lines[3].split()))

        # QL weights
        with open(self['out.tglf.QL_flux_spectrum'].filename, 'r') as f:
            QL_data = np.loadtxt(f, skiprows=4, unpack=True).reshape(nky, nm, ns, nfield, ntype)
        particle_QL = QL_data[:, :, :, :, 0]
        energy_QL = QL_data[:, :, :, :, 1]
        toroidal_stress_QL = QL_data[:, :, :, :, 2]
        parallel_stress_QL = QL_data[:, :, :, :, 3]
        exchange_QL = QL_data[:, :, :, :, 4]

        with open(self['out.tglf.spectral_shift'].filename, 'r') as f:
            kx0_e = np.loadtxt(f, skiprows=5, unpack=True)

        with open(self['out.tglf.scalar_saturation_parameters'].filename, 'r') as f:
            ave_p0, B_unit, R_unit, q_unit, SAT_geo0_out, kx_geo0_out = np.loadtxt(f, skiprows=5, unpack=True)

        R_unit = np.ones((21, 2)) * R_unit

        return sum_ky_spectrum(
            sat_rule_in=saturation_rule_name,
            ky_spect=ky_spect,
            gp=gammas,
            ave_p0=ave_p0,
            R_unit=R_unit,
            kx0_e=kx0_e,
            potential=potential,
            particle_QL=particle_QL,
            energy_QL=energy_QL,
            toroidal_stress_QL=toroidal_stress_QL,
            parallel_stress_QL=parallel_stress_QL,
            exchange_QL=exchange_QL,
            etg_fact=1.25,
            c0=32.48,
            c1=0.534,
            exp1=1.547,
            cx_cy=0.56,
            alpha_x=1.15,
            **self['input.tglf'],
        )


def intensity_desat(ky_spect, gradP, q, taus_2):
    '''
    Dummy Experimental SATuration rule

    :param ky_spect: poloidal wave number [nk]

    :param gradP: P_PRIME_LOC (pressure gradient - see tglf inputs: https://gacode.io/tglf/tglf_list.html)

    :param q: absolute value of safety factor (i.e. Q_LOC)

    :param taus_2: ratio of T_main_ion/ T_e
    '''
    intensity_desat = (0.0082 / abs(gradP) + 0.1 * log10(ky_spect)) / (ky_spect ** ((3.81 * q) / (q + taus_2)))
    intensity_desat = np.c_[intensity_desat, intensity_desat]
    return intensity_desat


def intensity_sat0(
    ky_spect,
    gp,
    ave_p0,
    R_unit,
    kx0_e,
    etg_fact,
    c0,
    c1,
    exp1,
    cx_cy,
    alpha_x,
    B_unit=1.0,
    as_1=1.0,
    zs_1=-1.0,
    taus_1=1.0,
    mass_2=1.0,
    alpha_quench=0,
):
    '''
    SAT0 template for modifying the TGLF SAT0 intensity function from [Staebler et al., Nuclear Fusion, 2013], still needs to be normalized (see below)

    nk --> number of elements in ky spectrum
    nm --> number of modes
    ns --> number of species
    nf --> number of fields (1: electrostatic, 2: electromagnetic parallel, 3:electromagnetic perpendicular)

    :param ky_spect: poloidal wave number [nk]

    :param gp: growth rates [nk, nm]

    :param ave_p0: scalar average pressure

    :param R_unit: scalar normalized major radius

    :param kx0_e: spectral shift of the radial wavenumber due to VEXB_SHEAR [nk]

    :param etg_fact: scalar TGLF calibration coefficient [1.25]

    :param c0: scalar TGLF calibration coefficient [32.48]

    :param c1: scalar TGLF calibration coefficient [0.534]

    :param exp1: scalar TGLF calibration coefficient [1.547]

    :param cx_cy: scalar TGLF calibration coefficient [0.56] (from Eq.13)

    :param alpha_x: scalar TGLF calibration coefficient [1.15] (from Eq.13)

    :param B_unit: scalar normalized magnetic field (set to 1.0)

    :param as_1: scalar ratio of electron density to itself (must be 1.0)

    :param zs_1: scalar ratio of electron charge to the first ion charge (very likely to be -1.0)

    :param taus_1: scalar ratio of electron temperature to itself (must be 1.0)

    :param mass_2: scalar ratio of first ion mass to itself (must be 1.0)

    :param alpha_quench: scalar if alpha_quench==0 and any element in spectral shift array is greater than zero (versus equal zero), apply spectral shift routine

    :return: intensity function [nk, nm]
    '''

    if as_1 != 1.0 or taus_1 != 1.0 or mass_2 != 1.0:
        raise ValueError('as_1, taus_1, mass_2 must be equal to 1.0')
    kp = np.c_[ky_spect, ky_spect]  # add column makes two column array for dependence on two modes
    # kx0_e = np.c_[kx0_e, kx0_e]
    ks = kp * np.sqrt(taus_1 * mass_2)
    pols = (ave_p0 / np.abs(as_1 * zs_1 * zs_1)) ** 2
    R_unit[R_unit == 0] = np.amax(R_unit)  # when EV solver does not converge, R_unit is 0.0, but should be nonzero and the same for all  ky
    wd0 = ks * np.sqrt(taus_1 / mass_2) / R_unit
    gnet = gp / wd0
    cond = (alpha_quench == 0) * (np.abs(kx0_e) > 0)
    notcond = (alpha_quench != 0) + (np.abs(kx0_e) == 0)
    # Calculate intensity_given using scalar TGLF calibration coefficients
    cnorm = c0 * pols
    cnorm = cnorm / ((ks ** etg_fact) * (ks > 1) + 1 * (ks <= 1))
    intensity_out = cnorm * (wd0 ** 2) * (gnet ** exp1 + c1 * gnet) / (kp ** 4)
    intensity_out = intensity_out / (((1 + cx_cy * kx0_e ** 2) ** 2) * cond + 1 * notcond)
    intensity_out = intensity_out / (((1 + (alpha_x * kx0_e) ** 4) ** 2) * cond + 1 * notcond)
    intensity_out = intensity_out / B_unit ** 2
    intensity_out = np.nan_to_num(intensity_out)
    intensity_out[
        intensity_out == 0
    ] = 10e10  # when EV solver did not converge, intensity and potential are zero; change zeros to large number before division
    return intensity_out  # Careful: This still needs to be normalized; next operation is intensity_out = intensity_out * potential(TGLF SAT0 output) / intensity_out(default SAT0 params)
    # See example with default parameters below and regression test 'omfit/regression/test_tglf_satpy.py'


def intensity_sat1(
    ky_spect,
    gp,
    kx0_e,
    taus_2,
    expsub=2.0,
    alpha_zf_in=1.0,
    kx_geo0_out=1.0,
    SAT_geo_out=1.0,
    bz1=0.0,
    bz2=0.0,
    zs_1=-1.0,
    zs_2=1.0,
    taus_1=1.0,
    mass_1=0.000272313,
    mass_2=1.0,
):
    '''
    TGLF SAT1 from [Staebler et al., 2016, PoP], takes both GYRO and TGLF outputs as inputs

    :param ky_spect: poloidal wavenumber [nk]

    :param gp: growth rates [nk, nm]

    :param kx0_e: spectral shift of the radial wavenumber due to VEXB_SHEAR [nk]

    :param expsub: scalar exponent in gammaeff calculation [2.0]

    :param alpha_zf_in: scalar switch for the zonal flow coupling coefficient [1.0]

    :param kx_geo_out: scalar switch for geometry [1.0]

    :param SAT_geo_out: scalar switch for geoemtry [1.0]

    :param bz1: scalar correction to zonal flow mixing term [0.0]

    :param bz2: scalar correction to zonal flow mixing term [0.0]

    :param zs_1: scalar ratio of electron charge to the first ion charge (very likely to be -1.0)

    :param zs_2: scalar ratio fo the first ion charge with itself [1.0]

    :param taus_1: scalar ratio of electron temperature to itself [1.0]

    :param taus_2: scalar ratio of first ion temperature to electron temperature

    :param mass_1: scalar ratio of electron to first ion [0.000272313]

    :param mass_2: scalar ratio of first ion mass to itself [1.0]

    '''

    nmodes_in = len(gp[0, :])
    nky = len(ky_spect)
    gammas1 = gp[:, 0]  # SAT1 uses the growth rates of the most unstable modes

    etg_streamer = 1.05
    kyetg = etg_streamer * abs(zs_2) / np.sqrt(taus_2 * mass_2)

    czf = abs(alpha_zf_in)
    small = 1.0e-10
    cz1 = 0.48 * czf
    cz2 = 1.0 * czf
    ax = 1.15
    ay = 0.56

    gamma_net = gammas1 / (1.0 + (ax * kx_geo0_out * kx0_e) ** 4)

    cky = 3.0
    sqcky = np.sqrt(cky)
    cnorm = 14.29

    kycut = 0.8 * abs(zs_2) / np.sqrt(taus_2 * mass_2)  # ITG/ETG-scale separation (for TEM scales see [Creely et al., PPCF, 2019])
    kyhigh = 0.15 / np.sqrt(taus_1 * mass_1)

    gammamax1 = gamma_net[0]
    kymax1 = ky_spect[0]
    testmax1 = gammamax1 / kymax1
    testmax2 = 0
    jmax1 = 0
    jmax2 = 0

    j1 = 0
    j2 = 0

    for j in range(1, nky):
        ky0 = ky_spect[j]
        if ky0 < kycut:
            j1 = j1 + 1
        if ky0 < kyhigh:
            j2 = j2 + 1
        test = gamma_net[j] / ky0
        if ky0 < kycut:
            if test > testmax1:
                testmax1 = test
                jmax1 = j
        if ky0 > kycut:
            if test > testmax2:
                testmax2 = test
                jmax2 = j
    # handle exceptions
    if j1 == nky - 1:
        j1 = nky - 2
    if jmax2 == 0:
        jmax2 = j2

    gammamax2 = gamma_net[jmax2]
    kymax2 = ky_spect[jmax2]
    gammamax1 = gamma_net[jmax1]
    kymax1 = ky_spect[jmax1]

    # Routine for better determination of gamma/ky peak
    if jmax1 > 0 and jmax1 < j1:
        f0 = gamma_net[jmax1 - 1] / ky_spect[jmax1 - 1]
        f1 = gamma_net[jmax1] / ky_spect[jmax1]
        f2 = gamma_net[jmax1 + 1] / ky_spect[jmax1 + 1]
        dky = ky_spect[jmax1 + 1] - ky_spect[jmax1 - 1]
        x0 = (ky_spect[jmax1] - ky_spect[jmax1 - 1]) / dky
        a = f0
        x02 = x0 * x0
        b = (f1 - f0 * (1 - x02) - f2 * x02) / (x0 - x02)
        c = f2 - f0 - b
        xmax = -b / (2.0 * c)
        if xmax > 1.0:
            kymax1 = ky_spect[jmax1 + 1]
            gammamax1 = f2 * kymax1
        elif xmax < 0.0:
            kymax1 = ky_spect[jmax1 - 1]
            gammamax1 = f0 * kymax1
        else:
            kymax1 = ky_spect[jmax1 - 1] + dky * xmax
            gammamax1 = (a + b * xmax + c * xmax * xmax) * kymax1

    vzf1 = gammamax1 / kymax1
    vzf2 = gammamax2 / kymax2
    dvzf = max(vzf2 - vzf1, 0.0)
    dgamma = dvzf * kymax1
    vzf3 = max(vzf1 - bz1 * dvzf, 0.0)
    vzf4 = max(vzf1 - bz2 * dvzf, 0.0)

    # include zonal flow effects on growth rate model:
    gamma_mix1 = np.zeros(nky)
    gamma = np.zeros(nky)

    for j in range(0, nky):
        gamma0 = gamma_net[j]
        if ky_spect[j] < kymax1:
            gamma[j] = max(gamma0 - cz1 * (kymax1 - ky_spect[j]) * vzf1, 0.0)
        else:
            gamma[j] = cz2 * gammamax1 + max(gamma0 - cz2 * vzf1 * ky_spect[j], 0.0)
        gamma_mix1[j] = gamma[j]

    # Mix over ky>kymax with integration weight
    mixnorm1 = np.zeros(nky)
    for j in range(jmax1 + 2, nky):
        gamma_ave = 0.0
        mixnorm1 = ky_spect[j] * (
            np.arctan(sqcky * (ky_spect[nky - 1] / ky_spect[j] - 1.0)) - np.arctan(sqcky * (ky_spect[jmax1 + 1] / ky_spect[j] - 1.0))
        )
        for i in range(jmax1 + 1, nky - 1):
            ky_1 = ky_spect[i]
            ky_2 = ky_spect[i + 1]
            mix1 = ky_spect[j] * (np.arctan(sqcky * (ky_2 / ky_spect[j] - 1.0)) - np.arctan(sqcky * (ky_1 / ky_spect[j] - 1.0)))
            delta = (gamma[i + 1] - gamma[i]) / (ky_2 - ky_1)
            mix2 = ky_spect[j] * mix1 + (ky_spect[j] * ky_spect[j] / (2.0 * sqcky)) * (
                np.log(cky * (ky_2 - ky_spect[j]) ** 2 + ky_spect[j] ** 2) - np.log(cky * (ky_1 - ky_spect[j]) ** 2 + ky_spect[j] ** 2)
            )
            gamma_ave = gamma_ave + (gamma[i] - ky_1 * delta) * mix1 + delta * mix2
        gamma_mix1[j] = gamma_ave / mixnorm1

    # intensity
    nmodes_in = 2
    field_spectrum_out = np.zeros((21, 2))
    for j in range(0, nky):
        gamma0 = gp[j, 0]
        ky0 = ky_spect[j]
        kx = kx0_e[j]
        for i in range(0, nmodes_in):
            gammaeff = 0.0
            if gamma0 > small:
                gammaeff = gamma_mix1[j] * (gp[j, i] / gamma0) ** expsub
            if ky0 > kyetg:
                gammaeff = gammaeff * np.sqrt(ky0 / kyetg)
            field_spectrum_out[j, i] = SAT_geo_out * (cnorm * gammaeff * gammaeff / ky0 ** 4) / (1.0 + ay * (kx_geo0_out * kx) ** 2) ** 2
    phinorm = field_spectrum_out
    return phinorm  # is the intensity


def flux_integrals(
    NM,
    NS,
    NF,
    i,
    ky,
    dky0,
    dky1,
    particle,
    energy,
    toroidal_stress,
    parallel_stress,
    exchange,
    particle_flux_out,
    energy_flux_out,
    stress_tor_out,
    stress_par_out,
    exchange_out,
    q_low_out,
    taus_1=1.0,
    mass_2=1.0,
):
    '''
    Compute the flux integrals
    '''
    for nm in range(NM):
        for ns in range(NS):
            for j in range(NF):
                particle_flux_out[nm][ns][j] += dky0 * (0 if i == 0 else particle[i - 1][nm][ns][j]) + dky1 * particle[i][nm][ns][j]
                energy_flux_out[nm][ns][j] += dky0 * (0 if i == 0 else energy[i - 1][nm][ns][j]) + dky1 * energy[i][nm][ns][j]
                stress_tor_out[nm][ns][j] += (
                    dky0 * (0 if i == 0 else toroidal_stress[i - 1][nm][ns][j]) + dky1 * toroidal_stress[i][nm][ns][j]
                )
                stress_par_out[nm][ns][j] += (
                    dky0 * (0 if i == 0 else parallel_stress[i - 1][nm][ns][j]) + dky1 * parallel_stress[i][nm][ns][j]
                )
                exchange_out[nm][ns][j] += dky0 * (0 if i == 0 else exchange[i - 1][nm][ns][j]) + dky1 * exchange[i][nm][ns][j]
            if ky * taus_1 * mass_2 <= 1:
                q_low_out[nm][ns] = energy_flux_out[nm][ns][0] + energy_flux_out[nm][ns][1]
    return particle_flux_out, energy_flux_out, stress_tor_out, stress_par_out, exchange_out, q_low_out


def sum_ky_spectrum(
    sat_rule_in,
    ky_spect,
    gp,
    ave_p0,
    R_unit,
    kx0_e,
    potential,
    particle_QL,
    energy_QL,
    toroidal_stress_QL,
    parallel_stress_QL,
    exchange_QL,
    etg_fact=1.25,
    c0=32.48,
    c1=0.534,
    exp1=1.547,
    cx_cy=0.56,
    alpha_x=1.15,
    **kw,
):
    '''
    Perform the sum over ky spectrum
    The inputs to this function should be already weighted by the intensity function

    nk --> number of elements in ky spectrum
    nm --> number of modes
    ns --> number of species
    nf --> number of fields (1: electrostatic, 2: electromagnetic parallel, 3:electromagnetic perpendicular)

    :param sat_rule_in:

    :param ky_spect: k_y spectrum [nk]

    :param gp: growth rates [nk, nm]

    :param ave_p0: scalar average pressure

    :param R_unit: scalar normalized major radius

    :param kx0_e: spectral shift of the radial wavenumber due to VEXB_SHEAR [nk]

    :param potential: input potential fluctuation spectrum  [nk, nm]

    :param particle_QL: input particle fluctuation spectrum [nk, nm, ns, nf]

    :param energy_QL: input energy fluctuation spectrum [nk, nm, ns, nf]

    :param toroidal_stress_QL: input toroidal_stress fluctuation spectrum [nk, nm, ns, nf]

    :param parallel_stress_QL: input parallel_stress fluctuation spectrum [nk, nm, ns, nf]

    :param exchange_QL: input exchange fluctuation spectrum [nk, nm, ns, nf]

    :param etg_fact: scalar TGLF SAT0 calibration coefficient [1.25]

    :param c0: scalar TGLF SAT0 calibration coefficient [32.48]

    :param c1: scalar TGLF SAT0 calibration coefficient [0.534]

    :param exp1: scalar TGLF SAT0 calibration coefficient [1.547]

    :param cx_cy: scalar TGLF SAT0 calibration coefficient [0.56] (from TGLF 2008 POP Eq.13)

    :param alpha_x: scalar TGLF SAT0 calibration coefficient [1.15] (from TGLF 2008 POP Eq.13)

    :param \**kw: any additional argument should follow the naming convention of the TGLF_inputs
            for SAT1 the TAUS_2 argument is needed

    :return: dictionary with summations over ky spectrum:
            * particle_flux_integral: [nm, ns, nf]
            * energy_flux_integral: [nm, ns, nf]
            * toroidal_stresses_integral: [nm, ns, nf]
            * parallel_stresses_integral: [nm, ns, nf]
            * exchange_flux_integral: [nm, ns, nf]
    '''
    phi_bar_sum_out = 0
    NM = len(energy_QL[0, :, 0, 0])  # get the number of modes
    NS = len(energy_QL[0, 0, :, 0])  # get the number of species
    NF = len(energy_QL[0, 0, 0, :])  # get the number of fields
    particle_flux_out = np.zeros((NM, NS, NF))
    energy_flux_out = np.zeros((NM, NS, NF))
    stress_tor_out = np.zeros((NM, NS, NF))
    stress_par_out = np.zeros((NM, NS, NF))
    exchange_out = np.zeros((NM, NS, NF))
    q_low_out = np.zeros((NM, NS))

    # Multiply QL weights with desired intensity
    if sat_rule_in in [0.0, 0, 'SAT0']:
        intensity_factor = (
            intensity_sat0(ky_spect, gp, ave_p0, R_unit, kx0_e, etg_fact, c0, c1, exp1, cx_cy, alpha_x)
            * potential
            / intensity_sat0(ky_spect, gp, ave_p0, R_unit, kx0_e, 1.25, 32.48, 0.534, 1.547, 0.56, 1.15)
        )
    elif sat_rule_in in [1.0, 1, 'SAT1']:
        intensity_factor = intensity_sat1(ky_spect, gp, kx0_e, kw['TAUS_2'])
    elif sat_rule_in in [-1.0, -1, 'DESAT']:
        intensity_factor = (
            intensity_desat(ky_spect, kw['P_PRIME_LOC'], kw['Q_LOC'], kw['TAUS_2'])
            * potential
            / intensity_sat0(ky_spect, gp, ave_p0, R_unit, kx0_e, 1.25, 32.48, 0.534, 1.547, 0.56, 1.15)
        )
    else:
        raise ValueError("sat_rule_in must be [0.0, 0, 'SAT0'] or [1.0, 1, 'SAT1']")

    shapes = [item.shape for item in [particle_QL, energy_QL, toroidal_stress_QL, parallel_stress_QL, exchange_QL] if item is not None][0]

    particle = np.zeros(shapes)
    energy = np.zeros(shapes)
    toroidal_stress = np.zeros(shapes)
    parallel_stress = np.zeros(shapes)
    exchange = np.zeros(shapes)

    for i in range(NS):  # iterate over the species
        for j in range(NF):  # iterate over the fields
            if particle_QL is not None:
                particle[:, :, i, j] = particle_QL[:, :, i, j] * intensity_factor
            if energy_QL is not None:
                energy[:, :, i, j] = energy_QL[:, :, i, j] * intensity_factor
            if toroidal_stress_QL is not None:
                toroidal_stress[:, :, i, j] = toroidal_stress_QL[:, :, i, j] * intensity_factor
            if parallel_stress_QL is not None:
                parallel_stress[:, :, i, j] = parallel_stress_QL[:, :, i, j] * intensity_factor
            if exchange_QL is not None:
                exchange[:, :, i, j] = exchange_QL[:, :, i, j] * intensity_factor

    dky0 = 0
    ky0 = 0
    for i in range(len(ky_spect)):
        ky = ky_spect[i]
        ky1 = ky
        if i == 0:
            dky1 = ky1
        else:
            dky = np.log(ky1 / ky0) / (ky1 - ky0)
            dky1 = ky1 * (1.0 - ky0 * dky)
            dky0 = ky0 * (ky1 * dky - 1.0)

        particle_flux_out, energy_flux_out, stress_tor_out, stress_par_out, exchange_out, q_low_out = flux_integrals(
            NM,
            NS,
            NF,
            i,
            ky,
            dky0,
            dky1,
            particle,
            energy,
            toroidal_stress,
            parallel_stress,
            exchange,
            particle_flux_out,
            energy_flux_out,
            stress_tor_out,
            stress_par_out,
            exchange_out,
            q_low_out,
        )
        ky0 = ky1
        results = {
            "particle_flux_integral": particle_flux_out,
            "energy_flux_integral": energy_flux_out,
            "toroidal_stresses_integral": stress_tor_out,
            "parallel_stresses_integral": stress_par_out,
            "exchange_flux_integral": exchange_out,
        }
    return results



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


from omfit_classes.omfit_ascii import OMFITascii

import numpy as np
from astropy.io import ascii as asciitable

__all__ = ['OMFITasciitable']


class OMFITasciitable(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with ascii tables files

    This class makes use of the asciitable Python module http://cxc.harvard.edu/contrib/asciitable/

    Files can have a header preceeding the talbe (e.g. comment) and will be stored in the ['header'] field

    NOTE: the ['data'] field is a np.recarray and the data in the tables columns can be retrieved with the names defined in the ['columns'] field

    :param filename: filename passed to OMFITascii class

    :param skipToLine: (integer) line to skip to before doing the parsing

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.addedDot = False
        self.skipToLine = kw.get('skipToLine', 0)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """
        self['header'] = ''
        self['columns'] = []
        self['data'] = np.array([])

        with open(self.filename, 'r') as f:
            lines = f.read()

        if not len(lines.strip()):
            return

        if 'd' in lines.lower():
            # convert FORTRAN exponentials (d,D) to (e,E)
            lines = re.sub(r'([0-9]+|[0-9]*(?:[0-9]\.|\.[0-9])[0-9]*)[dD]([\-\+]*[0-9]+)', r'\1e\2', lines)
            with open(self.filename, 'w') as f:
                f.write(lines)

        lines = lines.splitlines(True)
        self['header'] = ''.join(lines[: self.skipToLine])
        lines = lines[self.skipToLine :]

        if len(lines):
            for k in range(len(lines)):
                tmp = ''.join(lines[k:])
                try:
                    if asciitable.__name__ == 'asciitable':
                        self['data'] = asciitable.read(tmp)
                    else:
                        self['data'] = np.array(asciitable.read(tmp, guess=True)).view(np.recarray)
                    self.addedDot = False
                    break
                except Exception:
                    tmp = '. ' + tmp
                    try:
                        if asciitable.__name__ == 'asciitable':
                            self['data'] = asciitable.read(tmp)
                        else:
                            self['data'] = np.array(asciitable.read(tmp, guess=True)).view(np.recarray)
                        self.addedDot = True
                        break
                    except Exception as _excp:
                        pass
            if k == len(lines):
                raise _excp

            self['columns'] = list(self['data'].dtype.names)
            self['header'] = self['header'] + ''.join(lines[:k])[:-1]
        self.dynaLoad = False

    @dynaSave
    def save(self):
        """
        Method used to save the content of the object to the file specified in the .filename attribute

        :return: None
        """
        with open(self.filename, 'w') as f:
            if len(self['header']):
                f.write(self['header'] + '\n')

            if np.all(['col' in item for item in self['columns']]):
                if asciitable.__name__ == 'asciitable':
                    asciitable.write(self['data'], f, asciitable.NoHeader)
                else:
                    asciitable.write(self['data'], f, 'fixed_width_no_header', delimiter=' ')
            else:
                if asciitable.__name__ == 'asciitable':
                    asciitable.write(self['data'], f, asciitable.Basic)
                else:
                    asciitable.write(self['data'], f, 'fixed_width', delimiter=' ')


############################################
if '__main__' == __name__:
    test_classes_main_header()

    filename = 'omfit_asciitable_test.dat'
    data = OMFITasciitable(filename, fromString='a b c\n1 2 3')
    data.load()
    data.save()
